﻿/**
 * @license Copyright © 2013 onwards, Andrew Whewell
 * All rights reserved.
 *
 * Redistribution and use of this software in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 *    * Neither the name of the author nor the names of the program's contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OF THE SOFTWARE BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/**
 * @fileoverview A jQuery UI plugin that wraps date controls in the configuration screens.
 */

(function(VRS, $, /** option= */undefined)
{
    //region jQueryUIHelper
    VRS.jQueryUIHelper = VRS.jQueryUIHelper || {};

    /**
     * Returns the VRS.vrsOptionFieldDate attached to the jQuery element passed across.
     * @param {jQuery} jQueryElement
     * @returns {VRS.vrsOptionFieldDate}
     */
    VRS.jQueryUIHelper.getOptionFieldDatePlugin = function(jQueryElement) { return jQueryElement.data('vrsVrsOptionFieldDate'); };
    //endregion

    //region vrsOptionFieldNumeric
    /**
     * @namespace VRS.vrsOptionFieldDate
     */
    $.widget('vrs.vrsOptionFieldDate', {
        //region -- options
        options: {
            /** @type {VRS.OptionField} */ field: undefined,

            __nop: null
        },
        //endregion

        //region -- _create
        /**
         * Creates the UI for the widget.
         * @private
         */
        _create: function()
        {
            var field = this.options.field;
            /** @type {VRS.OptionDateSettings} */ var settings = field.getTypeSettings() || new VRS.OptionDateSettings(null);

            var input = this.element;
            var dateTimePickerOptions = {
                defaultDate:    settings.defaultDate,
                minDate:        settings.minDate,
                maxDate:        settings.maxDate,
                onSelect:       function() { input.change(); }
            };
            $.extend(dateTimePickerOptions, VRS.globalisation.getDatePickerOptions());

            input.change(function() {
                if(VRS.timeoutManager) VRS.timeoutManager.resetTimer();

                var val = input.datepicker('getDate');
                if(val !== null) {
                    if(settings.minDate && settings.minDate !== 0 && val < settings.minDate) val = settings.minDate;
                    if(settings.maxDate && settings.maxDate !== 0 && val > settings.maxDate) val = settings.maxDate;

                    field.setValue(val);
                    field.saveState();
                }
            });

            var label = null;
            if(field.getLabelKey()) {
                label = $('<label/>')
                    .text(VRS.globalisation.getText(field.getLabelKey()) + ':')
                    .insertBefore(this.element);
            }

            this.element
                .uniqueId()
                .datepicker(dateTimePickerOptions);
            field.applyInputClass(this.element);

            this.element.val($.datepicker.formatDate(dateTimePickerOptions.dateFormat, field.getValue(), dateTimePickerOptions));

            if(label) label.attr('for', this.element.attr('id'));
        },

        /**
         * Releases resources allocated to the plugin.
         * @private
         */
        _destroy: function()
        {
            var label = $('label[for="' + this.element.attr('id') + '"');
            label.remove();

            this.element.datepicker('destroy');
            this.element.off();
        },
        //endregion

        __nop: null
    });
    //endregion

    //region Register control type
    if(VRS.optionControlTypeBroker) VRS.optionControlTypeBroker.addControlTypeHandlerIfNotRegistered(
        VRS.optionControlTypes.date,
        function(optionField, parentJQ) {
            return $('<input/>')
                .attr('type', 'text')
                .appendTo(parentJQ)
                .vrsOptionFieldDate({
                    field: optionField
                });
        }
    );
    //endregion
}(window.VRS = window.VRS || {}, jQuery));